import queue
import logging
import threading

from memorious import settings
from memorious.logic.rate_limit import get_rate_limit
from memorious.logic.context import Context
from memorious.model.queue import Queue

log = logging.getLogger(__name__)


class TaskRunner(object):
    """A long running task runner that uses Redis as a task queue"""

    @classmethod
    def execute(cls, stage, state, data, next_allowed_exec_time=None):
        """Execute the operation, rate limiting allowing."""
        try:
            context = Context.from_state(state, stage)
            if context.crawler.disabled:
                return
            if context.stage.rate_limit:
                resource = "%s:%s" % (context.crawler.name, context.stage.name)
                rate_limit = get_rate_limit(
                    resource, limit=context.stage.rate_limit
                )
                if rate_limit.check():
                    context.execute(data)
                    rate_limit.update()
                else:
                    Queue.queue(stage, state, data)
            else:
                context.execute(data)
        except Exception:
            log.exception("Task failed to execute:")
        finally:
            Queue.task_done(context.crawler, stage, state)
            # If we don't have anymore tasks to execute, time to clean up.
            if not context.crawler.is_running:
                context.crawler.aggregate(context)

    @classmethod
    def process(cls, q):
        while True:
            item = q.get()
            if item is None:
                q.task_done()
                return
            try:
                cls.execute(*item)
            except Exception:
                log.exception("Task failed to execute:")
            q.task_done()

    @classmethod
    def run_sync(cls):
        for task in Queue.tasks(stop_on_timeout=True):
            cls.execute(*task)

    @classmethod
    def run(cls):
        log.info("Processing queue (%s threads)", settings.THREADS)
        q = queue.Queue(maxsize=settings.THREADS)
        threads = []
        for _ in range(settings.THREADS):
            t = threading.Thread(target=cls.process, args=(q,))
            t.daemon = True
            t.start()
            threads.append(t)

        while True:
            # This is to handle fakeredis. It terminates the iterator
            # in Queue.tasks() when it "temporarily" runs out of tasks
            # during the execution of the initial task. So we have to
            # restart Queue.tasks() after q.join() to make sure that
            # no child tasks were queued.
            tasks_processed = 0
            for item in Queue.tasks():
                tasks_processed += 1
                q.put(item)

            # block until all tasks are done
            q.join()
            if tasks_processed == 0:
                break

        # stop workers
        for _ in range(settings.THREADS):
            q.put(None)
        for t in threads:
            t.join()



import logging
from servicelayer.jobs import RateLimit
from memorious.core import conn

log = logging.getLogger(__name__)


def get_rate_limit(resource, limit=100, interval=60, unit=1):
    return RateLimit(conn, resource, limit=limit, interval=interval, unit=unit)



import logging

from servicelayer.jobs import JobStage, Progress, Job
from servicelayer.util import load_json, dump_json

from memorious.core import manager, conn

log = logging.getLogger(__name__)


class Queue(object):
    """Manage the execution of tasks in the system."""

    @classmethod
    def tasks(cls, stop_on_timeout=False):
        stages = list({str(stage) for _, stage in manager.stages})
        while True:
            job_stage, data, state = JobStage.get_stage_task(
                conn, stages, timeout=5
            )
            if not job_stage:
                if stop_on_timeout:
                    # Stop if timed out/ no task returned
                    return
                else:
                    continue
            yield (job_stage.stage, state, load_json(data))

    @classmethod
    def queue(cls, stage, state, data):
        crawler = state.get('crawler')
        job_stage = JobStage(conn, str(stage), state['run_id'], str(crawler))
        job_stage.queue_task(dump_json(data), state)

    @classmethod
    def size(cls, crawler):
        """Total operations pending for this crawler"""
        status = Progress.get_dataset_status(conn, str(crawler))
        return status.get('pending')

    @classmethod
    def is_running(cls, crawler):
        """Is the crawler currently running?"""
        if crawler.disabled:
            return False
        return cls.size(crawler) > 0

    @classmethod
    def flush(cls, crawler):
        Job.remove_dataset(conn, str(crawler))

    @classmethod
    def task_done(cls, crawler, stage, state):
        job_stage = JobStage(
            conn, str(stage), str(state['run_id']), str(crawler)
        )
        job_stage.task_done()
